#ifndef __DUNE_MMDG_COUPLEDDGPROBLEM_HH__
#define __DUNE_MMDG_COUPLEDDGPROBLEM_HH__

#include <dune/mmdg/problems/dgproblem.hh>

//abstract class providing an interface for a coupled flow problem in a
//fractured porous medium that is to be solved with a coupled discontinuous
//Galerkin scheme
template<class Coordinate, class Scalar = double>
class CoupledDGProblem : public DGProblem<Coordinate, Scalar>
{
public:
  static constexpr int dim = Coordinate::dimension;
  using Base = DGProblem<Coordinate, Scalar>;
  using Matrix = typename Base::Matrix;

  //the exact solution on the interface at position pos
  virtual Scalar exactInterfaceSolution (const Coordinate& pos) const
  {
    return Scalar(0.0);
  }

  //interface source term at position pos
  virtual Scalar qInterface (const Coordinate& pos) const
  {
    return Scalar(0.0);
  }

  //aperture d of the fracture at position pos
  virtual Scalar aperture (const Coordinate& pos) const
  {
    return Scalar(1.0);
  }

  //tangential gradient of the aperture d of the fracture at position pos
  virtual Coordinate gradAperture (const Coordinate& pos) const
  {
    return Coordinate(0.0);
  }

  virtual Scalar interfaceBoundary(const Coordinate& pos) const
  {
    return exactInterfaceSolution(pos);
  }

  //tangential permeability per aperture tensor of the interface at position pos
  virtual Matrix Kparallel (const Coordinate& pos) const
  {
    Matrix permeability(0.0);

    for (int i = 0; i < dim; i++)
    {
      permeability[i][i] = 1.0;
    }

    //return identity matrix
    return permeability;
  }

  //permeability per aperture of the fracture in normal direction
  //at position pos
  virtual Scalar Kperp (const Coordinate& pos) const
  {
    return Scalar(1.0);
  }

  //returns the recommended quadrature order to compute an integral
  //over x * qInterface(x)
  virtual int quadratureOrder_qInterface () const
  {
    return 10;//1;
  }

  //returns the recommended quadrature order to compute an integral
  //over  x * d^2 * interfaceBoundary(x)
  //and   x * d * interfaceBoundary(x) * (Kpar * grad(d))
  //and   d^2 * interfaceBoundary(x) * Kpar
  virtual int quadratureOrderInterfaceBoundary () const
  {
    return 10;//1;
  }

  //returns the recommended quadrature order to compute integrals
  //over  x^2 * (Kpar * grad(d) * grad(d))
  //and   x^2 * d^2
  //and   x^2 * d * (Kpar * grad(d))
  //and   x * d^2 * Kpar
  virtual int quadratureOrder_Kparallel () const
  {
    return 10;//2;
  }

  //returns the recommended quadrature order to compute an integral
  //over x * Kperp(x)
  virtual int quadratureOrder_Kperp () const
  {
    return 10;//1;
  }
};

#endif
