#ifndef __DUNE_MMDG_DGPROBLEM3_HH__
#define __DUNE_MMDG_DGPROBLEM3_HH__

#include <cmath>
#include <string>

#include <dune/mmdg/problems/dgproblem.hh>

//Poisson problem with right side q = -1
template<class Coordinate, class Scalar = double>
class DGProblem3 : public DGProblem<Coordinate, Scalar>
{
  public:
    static constexpr int dim = Coordinate::dimension;

    //the exact solution at position pos,
    //in 2D and 3D the exact solution is obtained by a Fourier series expansion
    //of the rectangular function
    Scalar exactSolution (const Coordinate& pos) const
    {
      if constexpr (dim == 1)
      {
        return 0.5 * pos[0] * (pos[0] - 1);
      }

      if constexpr (dim == 2) //Fourier series with cut-off
      {
        constexpr int cutoff = 31;
        double solution = 0.0;

        for (int m = 1; m <= cutoff; m = m + 2)
        {
          for (int n = 1; n <= cutoff; n = n + 2)
          {
            solution += sin(M_PI * m * pos[0]) * sin(M_PI * n * pos[1])
              / (m * n * (m * m + n * n));
          }
        }

        solution *= -16.0 / pow(M_PI, 4);
        return solution;
      }

      if constexpr (dim == 3) //Fourier series with cut-off
      {
        constexpr int cutoff = 21;
        double solution = 0.0;

        for (int k = 1; k <= cutoff; k = k + 2)
        {
          for (int l = 1; l <= cutoff; l = l + 2)
          {
            for (int m = 1; m <= cutoff; m = m + 2)
            {
              solution += sin(M_PI * k * pos[0]) * sin(M_PI * l * pos[1])
                * sin(M_PI * m* pos[2]) / (k * l * m * (k * k + l * l + m * m));
            }
          }
        }

        solution *= -64.0 / pow(M_PI, 5);
        return solution;
      }

      DUNE_THROW(Dune::Exception,
        "Exact Solution not implemented for dimension = " +
        std::to_string(dim) + "!");
    }

    //exact solution is implemented
    bool hasExactSolution() const
    {
      return true;
    };

    //source term
    Scalar q (const Coordinate& pos) const
    {
      return Scalar(-1.0);
    }

    //boundary condition at position pos
    virtual Scalar boundary (const Coordinate& pos) const
    {
      return Scalar(0.0);
    }
};

#endif
