import numpy as np
from os import mkdir, remove
from os.path import join, exists
from subprocess import call
from math import log


# writes a dgf file for a grid with N elements per directions and dimension dim
def writeDGF(N, dim):
     # open dgf file
    file = open(join("grids", "cube" + str(dim) + "d.dgf"), "w")

    file.write("DGF\n\nInterval\n")

    # write lower left corner
    for i in range(0, dim):
        file.write(str(0) + "\t")
    file.write("\n")

    # write upper right corner
    for i in range(0, dim):
        file.write(str(1) + "\t")
    file.write("\n")

    # write number of grid elements
    for i in range(0, dim):
        file.write(str(N) + "\t")
    file.write("\n#\n\nSimplex\n#\n\nGridparameter\noverlap\t0\n#")

    # close dgf file
    file.close()


# reads error, run time and maximum edge length, i.e. the output from the
# excuted discontinuous Galerkin scheme
def readData():
    file = open("convergenceData", "r")

    # read data
    error = float(file.readline())
    timeTotal = float(file.readline())
    numberOfElements = float(file.readline())

    file.close()

    # return data
    return np.array([error, timeTotal, numberOfElements])


# determines the parameter "fastEOC" from the ini file
def determineFastEOC ():
    # boolean that determines whether to exclude computationally expensive grids
    fastEOC = False

    # read ini file content
    with open("parameterDG.ini", "r") as file:
        lines = file.read().splitlines()

        for i in range(len(lines)):
            if (lines[i].startswith("fastEOC") and \
             lines[i].rstrip().lower().endswith("true")):
                fastEOC = True
                break

    return fastEOC


# === main ===

# create directory for the plots if necessary
if not exists("plots"):
    mkdir("plots")

# number of runs from which the run times are taken as mean value
numberOfMeans = 1

# numbers of grid elements per direction
numElemsPerDir_12 = np.unique(np.logspace(0, 2.7, 15).astype(int))
numElemsPerDir_3 = np.unique(np.logspace(0, 1.5, 8).astype(int))

if determineFastEOC(): # exclude computationally expensive grids
    numElemsPerDir_12 = numElemsPerDir_12[:-3]
    numElemsPerDir_3 = numElemsPerDir_3[:-2]
else:
    print("fastEOC = false. This may take some minutes.\n")

for dim in [1, 2, 3]:

    numElemsPerDir = numElemsPerDir_3 if (dim == 3) else numElemsPerDir_12

    # storage for errors, run times and largestEdgeLength
    # column 0: error
    # column 1: total run time
    # column 2: largest edge length
    # different rows correspond to different numbers of grid elements
    data = np.zeros((len(numElemsPerDir), 3))

    for i in range(len(numElemsPerDir)):
        # write dgf file
        writeDGF(numElemsPerDir[i], dim)

        # take mean value over numberOfMeans runs
        for j in range(numberOfMeans):
            # run executable
            call(join(".", "dg-" + str(dim) + "d"))

            # store data
            data[i, :] += readData()
        # end for j
    # end for i

    # divide by numberOfMeans
    data /= numberOfMeans

    # determine experimental order of convergence (EOC)
    EOC = np.empty(len(numElemsPerDir) - 1)

    for i in range(len(numElemsPerDir) - 1):
        if data[i, 0] != 0:
            EOC[i] = log(data[i, 0] / data[i+1, 0]) / \
                log(data[i, 2] / data[i+1, 2])
        else:
            EOC[i] = -1.0*float('inf')
    # end for i

    # print EOCs
    print("\n-------------------------------------------------------------\n")
    print("EOCs (" + str(dim) + "d):\t")
    print(EOC)
    print("\n=============================================================\n\n")

    # write data to file
    np.savetxt(join("plots", "dgAnalysis_" + str(dim) + "d"), data)
    np.savetxt(join("plots", "dgEOC_" + str(dim) + "d"), EOC)
# end for dim

# remove output files
remove("convergenceData")
